!****************************************************************************
!  DESCRIPTION...: Library for XML Export/Import
!  COPYRIGHT.....: Copyright 2003, 2013 SoftVelocity Inc. All rights reserved
!
!  Refactoring of cpxml.inc, cpxml.clw and cpxmlif.inc in 2013 by David Harms
!  with the express permission of SoftVelocity and the assistance
!  of The Descartes Systems Group Inc.
!****************************************************************************

                                            Member
                                            Map
                                                module('')
!===  DOM  ======
CreateDOMImplementation                             PROCEDURE, *System_XML_CenterPoint_DOMImplementation, name('_CreateDOMImplementation@0'), pascal, raw
DestroyDOMImplementation                            PROCEDURE(*System_XML_CenterPoint_DOMImplementation DOMImplementation), name('_DestroyDOMImplementation@4'), pascal, raw
CreateDocumentFromFile                              PROCEDURE(const *CSTRING uri, ULONG poolSize = 251), *System_XML_CenterPoint_Document, name('_CreateDocument@8'), pascal, raw
CreateDocumentFromXML                               PROCEDURE(const *CSTRING xml, ULONG poolSize = 251), *System_XML_CenterPoint_Document, name('_CreateDocument2@8'), pascal, raw
CreateDOMWriter                                     PROCEDURE,*System_XML_CenterPoint_DOMWriter, name('_CreateDOMWriter@0'), pascal, raw
DestroyDOMWriter                                    PROCEDURE(*System_XML_CenterPoint_DOMWriter DOMWriter), name('_DestroyDOMWriter@4'), pascal, raw
                                                end

                                            End



    Include('System_XML_Document.inc'),Once
    Include('System_XML_CString.inc'),Once
    Include('System_Logger.inc'),Once

log                                         System_Logger
CS                                          System_XML_CString
NullCString                                 cstring('')
                                            itemize, pre(Format)
AS_IS                                           EQUATE(1)
CANONICAL                                       EQUATE(2)
REFORMATTED                                     EQUATE(4)
                                            end

    PRAGMA('link(C%V%cpxml.lib)')

System_XML_Document.Construct               Procedure()
    code
    self.DocumentElement &= new System_XML_Element
    self.DocumentElement.SetNodeName('xml')
    self.DocumentElement.IsRootElement = true
    self.InstantiateXPath()
    
System_XML_Document.Destruct                Procedure()
    code
    self.DisposeXPath()
    dispose(self.DocumentElement)
    
System_XML_Document.AddElementToDom         procedure(System_XML_CenterPoint_Document DOMDocument,System_XML_Node modeledNode,System_XML_CenterPoint_Element parentDOMElement)
x                                               long
DOMElement                                      &System_XML_CenterPoint_Element
DOMAttr                                         &System_XML_CenterPoint_Attr
Attr                                            &System_XML_Attr
ChildModeledNode                                &System_XML_Node
    code
    !log.write('AddElementToDom receives node ' & modeledNode.GetNodeName()) 
    case modeledNode.NodeType
    of System_XML_NodeType:Element
        if modeledNode.IsRootElement
            DOMElement &= parentDOMElement
        else
            DOMElement &= DOMDocument.CreateElement(modeledNode.NodeName)        
        end
        ! Add any attributes
        loop x = 1 to modeledNode.Attributes.GetCount()
            Attr &= modeledNode.Attributes.GetAttr(x)
            if not Attr &= null
                DOMAttr &= DOMDocument.CreateAttribute(attr.Name)
                DOMAttr.SetValue(cs.GetCString(attr.GetValue()))
                DOMElement.SetAttributeNode(DOMAttr)
                DOMAttr.Release()
            end
        end
        loop x = 1 to modeledNode.ChildNodes.GetCount()
            ChildModeledNode &= modeledNode.ChildNodes.GetNode(x)
            if not ChildModeledNode &= null
                self.AddElementToDom(DOMDocument,ChildModeledNode,DOMElement)
            end
        end
        if modeledNode.IsRootElement = false
            parentDOMElement.AppendChild(DOMElement)
            DOMElement.Release()
        end
    end     
    
System_XML_Document.CreateChildNode         procedure(string parentName,string nodeName)!,*System_XML_Node,virtual
ChildModeledNode                                &System_XML_Node
    code
    ! This is the default node type - this method will usually be derived to return
    ! a node appropriate to the node name
    ChildModeledNode &= new System_XML_Node
    ChildModeledNode.SetNodeName(nodeName)
    return ChildModeledNode   
    
System_XML_Document.CreateCDATASection      PROCEDURE(STRING data)!,*System_XML_CDATASection
DataSection                                     &System_XML_CDataSection
    code
    DataSection &= new System_XML_CDataSection
    return DataSection

System_XML_Document.CreateComment           PROCEDURE(STRING text)!,*System_XML_Comment
Comment                                         &System_XML_Comment
    code
    Comment &= new System_XML_Comment
    Comment.SetText(text)
    return comment
    
System_XML_Document.CreateElement           PROCEDURE(STRING tagName)!,*System_XML_Element
Element                                         &System_XML_Element
    code
    Element &= new System_XML_Element
    Element.SetNodeName(tagName)
    return Element  
    
System_XML_Document.CreateNode              PROCEDURE(STRING tagName)!,*System_XML_Node
Node                                            &System_XML_Node
    code
    Node &= new System_XML_Node
    Node.SetNodeName(tagName)
    return Node  
    
System_XML_Document.CreateObjectGraph       procedure(System_XML_CenterPoint_Document document)!,long
Element                                         &System_XML_CenterPoint_Element
    code
    self.DocumentElement.Reset() ! Make sure there's no pre-existing data
    Element &= Document.GetDocumentElement()
    self.WalkDomTreeAndCreateObjects(self.DocumentElement,Element)  
    self.OnObjectGraphCreated()
    
System_XML_Document.CreateTextNode          PROCEDURE(STRING data)!,*System_XML_Text
text                                            &System_XML_Text
    code
    text &= new System_XML_Text
    text.SetText(data)
    return text
    
System_XML_Document.DisposeXPath                    procedure!,virtual
    code
    dispose(self.Query)

System_XML_Document.FromFile                procedure(string filename,bool BuildObjectGraph=true)!,long
poolSize                                        ulong(251)
Element                                         &System_XML_CenterPoint_Element
Document                                        &System_XML_CenterPoint_Document
    code
    Document &= CreateDocumentFromFile(cs.GetCString(filename), poolSize)
    if Document &= null 
        return Level:Fatal
    end
    if BuildObjectGraph
        self.CreateObjectGraph(Document)
    end
    return level:benign
    
System_XML_Document.FromString              procedure(string xml)!,long
poolSize                                        ulong(251)
DOMDocument                                     &System_XML_CenterPoint_Document
    code
    self.Reset()
    DOMDocument &= CreateDocumentFromXML(cs.GetCString(xml), poolSize)
    if DOMDocument &= null 
        !log.write('document is null - could not create from string')
        !self.Errors.AddError('Unable to load xml from string')
        return Level:Fatal
    end
    self.CreateObjectGraph(DOMDocument)
    return level:benign
    
    
System_XML_Document.GetDocumentElement      PROCEDURE!,*System_XML_Element
    code
    return self.DocumentElement
    
System_XML_Document.GetNodeText             procedure(System_XML_CenterPoint_Node domNode)!,string
textNode                                        &System_XML_CenterPoint_Node
    code
    if domNode &= null then return ''.
    domNode.Normalize()
    textNode &= domNode.getFirstChild()
    if (not textNode &= null)
        return textNode.getNodeValue()
    end
    return ''
    
System_XML_Document.InstantiateXPath                procedure!,virtual
    code
    self.Query &= new System_XML_XPathQuery
    self.XPath &= self.Query.System_XML_XPath

System_XML_Document.GetFirstNode            procedure(string nodeName)!,*System_XML_Node
    code
    if self.DocumentElement &= null then return null.
    return self.DocumentElement.GetFirstNode(nodeName)
    
System_XML_Document.OnObjectGraphCreated    procedure!,virtual
    code

System_XML_Document.SetDocumentElementName  procedure(string name)
    code
    self.DocumentElement.SetNodeName(name)
    
System_XML_Document.ToString                procedure!,string
DOMImplementation                               &System_XML_CenterPoint_DOMImplementation
DocumentType                                    &System_XML_CenterPoint_DocumentType
Document                                        &System_XML_CenterPoint_Document
Writer                                          &System_XML_CenterPoint_DomWriter
DocumentElement                                 &System_XML_CenterPoint_Element
Output                                          System_XML_CString
    code
    DOMImplementation &= CreateDomImplementation()
    DocumentType &= DOMImplementation.CreateDocumentType(NullCString,NullCString,NullCString)
    !const *CSTRING namespaceURI,const *CSTRING qualifiedName,*System_XML_CenterPoint_DocumentType doctype,ULONG poolSize = 251),*System_XML_CenterPoint_Document,name('_createDocument@20'),pascal,raw
    Document &= DOMImplementation.CreateDocument(NullCString,self.DocumentElement.NodeName,DocumentType)
    DocumentElement &= Document.GetDocumentElement()
    self.AddElementToDOM(Document,self.DocumentElement,DocumentElement)

    Writer &= CreateDOMWriter()
    Writer.setFormat(format:reformatted + format:canonical)
    Output.Set(Writer.WriteNode(Document))
    Document.Release()
    DestroyDomImplementation(DOMImplementation)
    return clip(Output.GetString())
    
System_XML_Document.WalkDomTreeAndCreateObjects     procedure(System_XML_Node modeledNode,System_XML_CenterPoint_Node domNode)
ChildDomNode                                            &System_XML_CenterPoint_Node
AttributeDomNode                                        &System_XML_CenterPoint_Node
ChildNodeList                                           &System_XML_CenterPoint_NodeList
ChildModeledNode                                        &System_XML_Node
AttributeMap                                            &System_XML_CenterPoint_NamedNodeMap
x                                                       long
Level                                                   long,static
Element                                                 &System_XML_CenterPoint_Element
NodeAddress                                             long
!t                                                       &System_XML_Element
IsElement                                               bool
    CODE
    modeledNode.SetNodeName(domNode.GetNodeName())
    modeledNode.SetNodeValue(self.GetNodeText(domNode))
    level += 1
    if not self.XPath &= NULL
        self.XPath.OnNewNode(level,modeledNode)
    end 
    ! Extract attributes
    AttributeMap &= domNode.getAttributes()
    if not AttributeMap &= null
        loop x = 0 to AttributeMap.getLength() - 1
            AttributeDomNode &= AttributeMap.item(x)
            modeledNode.AddAttribute(AttributeDomNode.getNodeName(),AttributeDomNode.getNodeValue())
        end
    end
    AttributeMap.Release()
    ! Add children
    ChildNodeList &= domNode.getChildNodes()
    if not ChildNodeList &= null
        loop x = 0 to ChildNodeList.getLength() - 1
            ChildDomNode &= ChildNodeList.Item(x)
            if not ChildDomNode &= null 
                case ChildDomNode.GetNodeType()
                of System_XML_NodeType:ELEMENT
                    ChildModeledNode &= self.CreateChildNode(domNode.GetNodeName(),ChildDomNode.GetNodeName())
                    !ChildModeledNode &= new System_XML_Node ! self.CreateNode(ChildDomNode.GetNodeName())
                    !ChildModeledNode.SetNodeName(ChildDomNode.GetNodeName())
                    ChildModeledNode.NodeType = System_XML_NodeType:Element
                    modeledNode.ChildNodes.AddNode(ChildModeledNode)
                    self.WalkDomTreeAndCreateObjects(ChildModeledNode,ChildDomNode)
                end
            end
        end
    end
    ChildNodeList.release()    
    level -= 1